# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Re-usable styles for the Copycat UI."""

import mesop as me


DEFAULT_ROW_AND_COLUMN_GAP = 16
HEADER_SECTION_STYLE = dict(display="flex", gap=5, align_items="center")
ROUNDED_BOX_SECTION_STYLE = dict(
    border_radius=16,
    padding=me.Padding.all(24),
    border=me.Border.all(
        me.BorderSide(
            width=1,
            style="solid",
            color=me.theme_var("outline-variant"),
        )
    ),
)
ROUNDED_BOX_SECTION_HEADER_TYPE = "headline-5"

DEFAULT_BORDER_STYLE = me.BorderSide(
    width=1,
    style="solid",
    color=me.theme_var("outline-variant"),
)
